package org.eso.ohs.evm;

/**
 * This class is the primary interface used by intergrating applications
 * wishing to use the EVM module.
 *
 * The class will initialize itssefl with an Implementing object that is
 * either a ScriptEVMImpl or a SummyEVMImpl, depending whether the scripting
 * package (org.eso.ohs.scripting) is available in the classpath or not.
 *
 * This class (EVMExecutor) and the classes implementing EVMImpl comprise
 * the two key components of the Bridge pattern.
 *
 * NOTE: the timeout values are not used/not implemented
 *
 * @author Nick Kornweibel
 */

import java.io.File;

import org.eso.ohs.dfs.CalibrationBlock;



public class EVMExecutor {

    /**
     * The implementor object used by this class (see Bridge pattern)
     */
    private EVMImpl evmImpl_ = null;

    /**
     * The status of the EVMExecutor - "ok" if initialized, else an
     * error condition.
     */
    private String evmStatus_;

    /**
     * The name of the scripting package implementation of EVMImpl
     * It is public so that is may be set to another class name
     * if required (also for testing purposes.
     */
    public static String scriptingEVMImplClassName 
            = "org.eso.ohs.scripting.ScriptingEVMImpl";
    

    /**
     * Construstor expects script driver names to be loaded and a
     * script finder to be used to locate scripts
     * Successful initialization is determined with isInitialized()
     *
     * @param array of script drivers to be loaded
     * @param script finder object for locating scripts
     */

    public EVMExecutor(String[] scriptDrivers, EVMScriptFinder sf) {
        String error = "unknown error";
        try {
            // Class.forName(className, true, null);
            // for 1.0.8
            Class evmImpClass = Class.forName(scriptingEVMImplClassName);
            if (evmImpClass != null) {
                evmImpl_ = (EVMImpl)evmImpClass.newInstance();
                evmImpl_.initImpl(scriptDrivers, sf);
            } else {
                error = "EVM Implementation Class not found";
            }
        } catch (EVMException e) {
            error = e.toString();
            evmImpl_ = null;
        } catch (ClassNotFoundException e) {
            error = e.toString();
            evmImpl_ = null;
        } catch (IllegalAccessException e) {
            error = e.toString();
            evmImpl_ = null;
        } catch (InstantiationException e) {
            error = e.toString();
            evmImpl_ = null;
        }
        if (evmImpl_ == null) {
            // failed to load the scripting package/script drivers
            // so use the Dummy EVMImpl...
            System.out.println("!!!!!!!!!!!!!! Error : " + error);            
            disable(error);
        } else {
            evmStatus_ = new String("ok");
        }
    }

    /**
     * Used to test if the EVMExecutor is initialized with all requested
     * script drivers and a script finder object
     * If not initialized, a DummyEVMImpl object is used as the implementor.
     * @return true if scripting package was loaded and object was
     * initialized error-free
     */

    public boolean isInitialized() {
        return (evmImpl_ != null && !(evmImpl_ instanceof DummyEVMImpl));
    }

    /**
     * Verify the array of OBs. Messages are reported by adding a 
     * EVMMessageListener prior to verification
     * @param array of OBs to be verified
     * @throws EVMException if there are errors finding or executing the
     * verification scripts
     */

    public synchronized void verify(CalibrationBlock[] obs) throws EVMException {
        evmImpl_.verifyImpl(obs);
    }
    
    /**
     * Set the exposure/execution time.
     * @param obs
     * @throws EVMException
     */
    
    public synchronized void calculateETCS(CalibrationBlock [] obs) throws EVMException {        
        evmImpl_.calculateETCSImpl(obs);   
    }

    /**
     * Set the script timeout (maximum time a script may execute)
     * Note: not implemented.
     * @param the timeout in milliseconds, 0L for no timeout
     */

    public void setScriptTimeout(long millis) {
        evmImpl_.setScriptTimeoutImpl(millis);
    }

    /**
     * Get the script timeout value in milliseconds
     * Note: timeouts are not implemented
     * @return the timeout value in milliseconds
     */

    public long getScriptTimeout() {
        return evmImpl_.getScriptTimeoutImpl();
    }

    /**
     * Get the set of script file extensions supported
     * @return string array of file extensions of supported scripts
     */

    public String[] getSupportedScriptExtensions() {
         return evmImpl_.getSupportedScriptExtensionsImpl();
    }


    /**
     * Run a script
     * @param the file path of the script
     * @throws if there are problems executing the script
     */

    public synchronized void runScript(File filePath) throws EVMException {
        evmImpl_.runScriptImpl(filePath);
    }

    /**
     * Register a listener for script messages
     * @param the listener to be registered
     */

    public void addEVMMessageListener(EVMMessageListener m) {
        evmImpl_.addEVMMessageListenerImpl(m);
    }

    /**
     * De-register a listener
     * @param the listener to be de-registered
     */

    public void removeEVMMessageListener(EVMMessageListener m) {
        evmImpl_.removeEVMMessageListenerImpl(m);
    }

    /**
     * Replace the implementation object with a Dummy (one-way operation)
     * Used by specializations of EVMExecutor
     * @param the reason for disabling the EVM
     */

    protected void disable(String reason) {
        evmImpl_ = new DummyEVMImpl();
        evmStatus_ = reason;
    }

    public String getStatusString() {
        return evmStatus_;
    }
}
