package org.eso.ohs.persistence;

import java.io.Serializable;
import java.util.Enumeration;
import java.util.NoSuchElementException;
import java.util.Vector;

import org.eso.ohs.dfs.StorableObject;

/**
 * A class to simulate a directory hierarchy.  It makes no
 * changes to the real filesystem.  The intention is merely to
 * provide objects to describe the directory containing an
 * object.  These objects are mainly used by the StorageManagers
 * to specify the directory for a read/write operation.
 * 
 * <p>
 * Because of this correspondence, the terms 'node' and 'directory'
 * are used interchangeably here.
 * 
 * <p>
 * $Id: DirectoryNode.java,v 1.10 2003/07/03 08:49:49 ddorigo Exp $
 * <br>
 * @author Bob Kemp
 * 
 * @see ObjectManager
 * @see StorageManager
 * @see LocalDirMgr
 */

public class DirectoryNode implements Serializable {

    /** Version id for object serialization */
    static final long serialVersionUID = 93976372643102550L;

    private StorableObject sobj_;

    /** The id of this directory */
    private long id_;

    /** Parent of this directory */
    private DirectoryNode parent_;

    /** Subdirectories of this directory */
    private DirectoryNode[] subDirs_;


    /**
     * Construct a node with a given name and no parent, ie a root node.
     * @param name name of this directory node <em>(must be non-null)</em>
     */
    public DirectoryNode(long id) {
	this(null, id);
    }

    /**
     * Construct a node with a given name and parent.
     * @param parent Parent of the directory node to be created
     * @param name name of this directory node <em>(must be non-null)</em>
     */
    public DirectoryNode(DirectoryNode parent, long id) {
	parent_ = parent;
	id_     = id;
    }


    /**
     * Returns true if this node has a parent node, ie is not the root.
     * 
     * @return true if this is not the root node, false otherwise.
     */
    public boolean hasParent() {
	return (parent_ != null);
    }


    /**
     * Returns a reference to the parent node.
     * 
     * @return Reference to the parent, or null if this is a root node.
     */
    public DirectoryNode getParent() {
	return parent_;
    }

    /**
     * Returns the storable object describing this directory.
     * 
     * @return a StorableObject
     */
    public StorableObject getObject() {
	return sobj_;
    }

    /**
     * Sets the storable object describing this directory.
     * 
     * @param newValue a StorableObject
     */
    public void setObject(StorableObject newValue) {
	sobj_ = newValue;
    }

    /**
     * Returns the user's descriptive name of this node, i.e.&nbsp;of
     * this directory.
     * 
     * @return a string containing the name of this node.
     */
    public String getDescription() {
	return sobj_ == null ? "" : sobj_.getName();
    }

    /**
     * Sets the user's descriptive name of this node, i.e.&nbsp;of
     * this directory.
     * 
     * @param newValue a string containing the new description of this node.
     */
    public void setDescription(String newValue) {
	if (sobj_ != null) {
	    sobj_.setName(newValue);
	}
    }

    /**
     * Returns the id of this node, i.e.&nbsp;of this directory.
     * 
     * @return the id of this node.
     */
    public long getId() {
	return id_;
    }


    /**
     * Returns true if the child directory with the given id exists.
     * 
     * @param id	the id of the child directory.
     */
    public boolean hasSubdir(long id) {
	return getDirectory(id) != null;
    }

    /**
     * Returns the child directory with the given id, or null if
     * it doesn't exist.
     * 
     * @param id	the id of the child directory.
     */
    public DirectoryNode getDirectory(long id) {
	DirectoryNode child = null;

	if (subDirs_ != null) {
	    for (int j=0; j < subDirs_.length; j++) {
		if (subDirs_[j] != null && id == subDirs_[j].id_) {
		    child = subDirs_[j];
		    break;
		}
	    }
	}

	return child;
    }


    /**
     * Returns the number of child nodes for this DirectoryNode.
     * 
     * @return	a non-negative integer
     */
    public int getChildCount() {
	int childCount=0;

	if (subDirs_ != null) {
	    for (int j=0; j < subDirs_.length; j++) {
		if (subDirs_[j] != null)
		    childCount++;
	    }
	}

	return childCount;
    }


    /**
     * Returns an array of the names of each immediate subdirectory.
     * 
     * @return an array of subdirectory names
     */
    public DirectoryNode[] getDirectories() {
	if (subDirs_ == null) {
	    return new DirectoryNode[] {};
	}

	int childCount=getChildCount();
	DirectoryNode[] children = new DirectoryNode[childCount];

	for (int j=0, k=0; j < subDirs_.length; j++) {
	    if (subDirs_[j] != null) {
		children[k++] = subDirs_[j];
	    }
	}

	return children;
    }

    /**
     * Returns an array of the DirectoryNodes between this node and
     * the root, inclusive.  We view this node as an ancestor of itself.
     * 
     * @return an array of ancestors
     */
    public DirectoryNode[] ancestors() {
	Vector ancestorVec = new Vector();
	DirectoryNode node;

	for (node=this; node != null; node=node.getParent()) {
	    ancestorVec.insertElementAt(node, 0);
	}

	DirectoryNode[] result = new DirectoryNode[ancestorVec.size()];
	ancestorVec.copyInto(result);

	return result;
    }

    /**
     * Returns true iff this directory is empty.
     * <p><b>NOTE</b>: takes no account of <em>files</em> within
     *	this directory, only subdirectories.  This will need to
     *	be extended once summary info is available.
     * 
     * @return true if the directory contains no subdirectories.
     */
    public boolean isEmpty() {
	if (subDirs_ != null) {
	    for (int j=0; j < subDirs_.length; j++) {
		if (subDirs_[j] != null) {
		    return false;
		}
	    }
	}
	return true;
    }


    /**
     * Attempts to create a subdirectory and returns true iff it
     * was successful.  Currently, this will only fail if the directory
     * already exists.
     * 
     * @return true if the directory was successfully created.
     */
    public boolean mkdir(long id) {
	return insert(new DirectoryNode(this, id));
    }

    private boolean insert(DirectoryNode newChild) {

	if (subDirs_ == null) {
	    subDirs_ = new DirectoryNode[] { null };
	}

	for (int j=0; j < subDirs_.length; j++) {
	    if (subDirs_[j] == null) {	
		subDirs_[j] = newChild;
		return true;
	    }
	}

		// No free slots -- increase space by 2 to leave
		//  a little extra room for expansion.
	DirectoryNode[] newDirs = new DirectoryNode[subDirs_.length+2];
	System.arraycopy(subDirs_, 0, newDirs, 0, subDirs_.length);
	subDirs_ = newDirs;
	subDirs_[subDirs_.length-1] = newChild;
	return true;
    }

    /**
     * Moves a directory within the DirectoryNode hierarchy.
     * 
     * @param toNode	the new parent directory
     */
    public boolean moveDirectory(DirectoryNode toNode) {
	DirectoryNode dir;

	for (int j=0; j < parent_.subDirs_.length; j++) {
	    if (parent_.subDirs_[j] == this) {
		if (toNode.insert(this)) {
		    parent_.subDirs_[j] = null;
		    parent_ = toNode;
		}
		return true;
	    }
	}
	return false;
    }

    /**
     * Attempts to remove a subdirectory and returns true iff it
     * was successful.  Currently, this will only fail if the
     * subdirectory doesn't exist or contains subdirectories of
     * its own.
     * 
     * @return true if the directory was successfully removed.
     */
    public boolean rmdir(long id) {
	DirectoryNode dir;

	for (int j=0; j < subDirs_.length; j++) {
	    dir = subDirs_[j];
	    if (dir != null && id == dir.id_) {
		if ( ! dir.isEmpty()) {
		    return false;
		}
		    // Flag to signify dead node
		subDirs_[j]       = null;
		return true;
	    }
	}

	return false;
    }


    /**
     * Returns a string representation of this object and its fields.
     * 
     * @return a string representation.
     */
    public String toString() {
	StringBuffer sb = new StringBuffer();

	sb.append("DirectoryNode[");
	sb.append("id=");
	sb.append(id_);

	if (parent_ != null) {
	    sb.append(", parent=");
	    sb.append(parent_.id_);
	}

	if (subDirs_ != null) {
	    for (int j=0; j < subDirs_.length; j++) {
		if (subDirs_[j] != null) {
		    sb.append(",<");
		    sb.append(j);
		    sb.append(">=");
		    sb.append(subDirs_[j].id_);
		}
	    }
	}

	sb.append("]");

	return sb.toString();
    }

    /**
     * Returns a Enumeration object that will successively return
     * this node and all of its descendants, ie child nodes, grandchild
     * nodes, etc.
     * 
     * @return a java.util.Enumeration object.
     */
    public Enumeration nodes() {
	return new Traversal(this);
    }


    /**
     * A utility class used to implement Enumeration over DirectoryNode trees.
     */
    private class Traversal implements Enumeration {

	/** Enumerator of subtrees */
	Enumeration enum_;

	/** Reference to the subtree (node+children) we are enumerating over */
	DirectoryNode node_;

	/** Index of our place in the child subtree array. */
	int index_;

	/**
	 * Construct an Enumeration object for the given node.
	 * 
	 * @param node the subtree over which to iterate.
	 */
	public Traversal(DirectoryNode node) {
	    enum_         = null;
	    node_         = node;
	    index_        = -1;
	}

	/**
	 * Return true iff there are more elements to be enumerated.
	 * 
	 * @return true if there are more elements to be enumerated.
	 */
	public boolean hasMoreElements() {
	    if (index_ < 0) {
		return true;
	    }

	    if (enum_ != null && enum_.hasMoreElements()) {
		return true;
	    }

	    if (node_.subDirs_ != null) {
		for ( ; index_ < node_.subDirs_.length ; index_++) {
		    if (node_.subDirs_[index_] != null) {
			enum_ = new Traversal(node_.subDirs_[index_]);
			index_++;
			return true;
		    }
		}
	    }

	    return false;
	}

	/**
	 * Returns the next object in the enumeration sequence or
	 * throws a NoSuchElementException exception if there is none.
	 * 
	 * @return The next node in the subtree.
	 * @exception NoSuchElementException if there are no more elements
	 * 					left to return.
	 */
	public Object nextElement() {
	    if (index_ < 0) {
		index_ = 0;
		return node_;
	    }

	    if ( ! hasMoreElements()) {
		throw new NoSuchElementException("DirectoryNode");
	    }
	    return enum_.nextElement();
	}
    }

}










