package org.eso.ohs.persistence;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.util.Enumeration;
import java.util.Stack;

import org.apache.log4j.Logger;
import org.eso.ohs.core.utilities.ObjectIOException;
import org.eso.ohs.dfs.BusinessObject;
import org.eso.ohs.dfs.FindingChart;
import org.eso.ohs.dfs.ObservingRun;
import org.eso.ohs.dfs.Readme;
import org.eso.ohs.dfs.Summary;
import org.eso.ohs.instruments.TemplateVersionInfo;

/**
 * The storage manager class for files stored in the local filesystem.
 * A concrete subclass of StorageManager.
 * 
 * <p>
 * $Id: LocalDirMgr.java,v 1.57 2004/12/14 17:47:35 tcanavan Exp $
 * @author Bob Kemp
 * <br>ChangeLog: cmavros 22Apr02 1195/P2PP Validation of OBs through P2PP AppServer.
 * 
 * @see StorageManager
 * @see ObjectManager
 */
public class LocalDirMgr implements StorageManager {

    private static Logger stdlog_ = Logger.getLogger(LocalDirMgr.class);

    /** Filename suffix for a Directory */
    public static final String DIRNAME_SUFFIX = "dir";

    /**
     * Temporary additional filename suffix for a new version of
     * a file before it is renamed to the correct value.
     * 
     * @see #recoverLocalStorage
     */
    public static final String NEWFILE_SUFFIX = "new";

    /**
     * Additional filename suffix for an old version of a file
     * so that the new version can be moved into place.
     * 
     * @see #recoverLocalStorage
     */
    public static final String OLDFILE_SUFFIX = "old";

    /* The time (in secs) when this object was created.  For getNewId() */
    final long creationTime_ = System.currentTimeMillis() / 10000;

    /** Private directory containing all user subtrees for local BO storage */
    private File localStoreRoot_;

    /**
     * Absolute path of the private subtree in which files should be saved.
     * Access is "protected" so that unit tests can set it.
     */
    protected String subtreeDir_;

    /** java.io.File object representing the subtreeDir_ directory */
    private File subtreeDirFile_;

    /** DirectoryNode object representing the subtreeDir_ directory */
    private DirectoryNode rootNode_;

    /** A counter used to set newly-created StorableObject's id value */
    private int idCounter = 0;

    /** a flag which indicates, whether the mkdir notification already has been
     *  done
     */
    private static boolean notify_mkdir_pop_up = false;

    /** UserId of the currently logged-in user */
    private int userId_;

    /**
     * Construct a storage manager for the given directory subtree.
     * This subtree should be in an area such that the user will not
     * alter the contents of the hierarchy.
     * 
     * @param localRootDir  root of the directory subtree containing
     *			    each user's local storage (aka cache).
     * 
     * @exception ObjectIOException
     *			if there was an I/O error while constructing the object
     */
    public LocalDirMgr(String localRootDir, int userId)
        throws ObjectIOException {
        localStoreRoot_ = new File(localRootDir);

        if (!localStoreRoot_.exists()) {
            notify_mkdir_pop_up = true;
            localStoreRoot_.mkdirs();
        }

        setUserId(userId);
    }

    public static boolean cacheDirsCreated() {
        return notify_mkdir_pop_up;
    }

    /**
     * Sets the id of the current user.
     * 
     * @param userId	the id of the end-user
     * 
     * @exception ObjectIOException
     *			if there was an I/O error while setting the userId
     */
    public void setUserId(int userId) throws ObjectIOException {
        if (userId == userId_ && subtreeDirFile_ != null) {
            return;
        }

        userId_ = userId;

        subtreeDirFile_ = new File(localStoreRoot_, userId_ + ".dir");
        rootNode_ = new DirectoryNode(-1);
        subtreeDir_ = subtreeDirFile_.getAbsolutePath();

        if (!Config.getCfg().isOperationsModeId(userId_)) {
            if (!subtreeDirFile_.exists()) {
                subtreeDirFile_.mkdirs();
            }
            if (!subtreeDirFile_.exists() || !subtreeDirFile_.isDirectory()) {
                throw new ObjectIOException(
                    "Cannot create local cache directory:\n" + subtreeDir_);
            }

            if (!subtreeDirFile_.canWrite()) {
                throw new ObjectIOException(
                    "Cannot write to local cache directory:\n" + subtreeDir_);
            }
        }

        initialise();
    }

    /**
     * Gets the id of the current user.
     * 
     * @return	the id of the end-user
     */
    public int getUserId() {
        return userId_;
    }

    /**
     * The root node of the directory tree.
     * 
     * @return a representation of the root of the directory tree.
     */
    public DirectoryNode getRoot() {
        return rootNode_;
    }

    /**
     * Verify that the TemplateSignature files used by the StorageManager
     * are the same version as those specified (and used by the client)
     * 
     * @param tsList	array of version info objects for each template
     *			used in an OB
     * 
     * @return a list of the names of TemplateSignatures that are out of date.
     */
    public String[] verifyTemplateSignatures(TemplateVersionInfo[] tsList) {
        // We're local, so by definition the templates are
        //  sufficiently uptodate.
        return new String[] {
        };
    }

    /**
     * Get the type of media maintained by this StorageManager instance.
     * 
     * @return an instance of the Media class.
     */
    public Media getMedia() {
        return Media.LOCAL;
    }

    /**
     * Calculate a unique id for local storage. 
     * 
     * @return Unique id
     */
    public synchronized long getNewId(Class objType, long baseId) {
        Config cfg = Config.getCfg();

        /*
        stdlog_.debug("GET_NEW_ID: objType = " + objType
        		    + ", baseId = " + baseId);
        */
        if (baseId != 0) {
            return cfg.tableToUniqueId(cfg.uniqueToTableId(baseId), objType);
        }

        long id = - (10 * creationTime_ + (++idCounter));

        return cfg.tableToUniqueId(id, objType);
    }

    /**
     * Authenticates the end-user.
     * 
     * @param password	password for the current user
     * 
     * @return	true if user was successfully authenticated, false otherwise
     */
    public boolean loginUser(String password) {
        return true;
    }

    /**
     * Returns a DirectoryNode representing the path within the
     * storage medium of an object file with the given id and type.
     * 
     * @param id	the id of the StorableObject's file.
     * @param objType	the type of stored StorableObject.
     * @return the directory in which to find the object.
     * @exception ObjectNotFoundException
     * 			if the file is not found.
     */
    public DirectoryNode find(long id, Class objType)
        throws ObjectNotFoundException {
        Long key = new Long(id);
        Enumeration enum;
        DirectoryNode node;
        File file;

        enum = getRoot().nodes();
        while (enum.hasMoreElements()) {
            node = (DirectoryNode) enum.nextElement();
            file = new Filename(node, id, objType).getFile();
            if (file.exists()) {
                return node;
            }
        }

        /** @cmavros 22-10-01 Add check for value less than zero */
        if (id < 0)
            throw new ObjectNotFoundException("no such object: 0");
        else
            throw new ObjectNotFoundException("no such object: " + id);
    }

    /**
     * Returns a DirectoryNode representing the path within the
     * storage medium of an object file with the given id and type.
     * 
     * @param id    the id of the StorableObject's file.
     * @param objType   the type of stored StorableObject.
     * @return the directory in which to find the object.
     * @exception ObjectNotFoundException
     *          if the file is not found.
     */
    public String locateRunPath(long id, Class objType)
        throws ObjectNotFoundException {
        Long key = new Long(id);
        Enumeration enum;
        DirectoryNode node;
        File file;

        enum = getRoot().nodes();
        while (enum.hasMoreElements()) {
            node = (DirectoryNode) enum.nextElement();
            file = new Filename(node, id, objType).getFile();
            if (file.exists()) {
                return file.getParentFile().getAbsolutePath();
            }
        }

        /** @cmavros 22-10-01 Add check for value less than zero */
        if (id < 0)
            throw new ObjectNotFoundException("no such object: 0");
        else
            throw new ObjectNotFoundException("no such object: " + id);
    }

    /**
     * Returns an array of objects representing the contents of an
     * object file with the given id and type from the chosen directory.
     * 
     * @param node	the directory containing the file.
     * @param id	the id of the object file.
     * @param objType	the type of stored object.
     * 
     * @return the array of objects and object references,
     * 			as for StorageManager.read().
     * 
     * @exception ObjectIOException
     * 			if an I/O error occurred while reading the file
     * 
     * @see #write
     * @see #update
     */
    public UnpackedStorableObject read(
        DirectoryNode node,
        long id,
        Class objType)
        throws ObjectIOException {
        File file = new Filename(node, id, objType).getFile();
        InputStream is;
        ObjectInputStream ois = null;
        UnpackedStorableObject obj;

        try {
            is = new FileInputStream(file.getAbsolutePath());
            ois = new ObjectInputStream(new BufferedInputStream(is));

            return (UnpackedStorableObject) ois.readObject();
        } catch (IOException e) {
            throw new ObjectIOException(e);
        } catch (ClassNotFoundException e) {
            throw new ObjectIOException(e);
        } finally {
            try {
                // It's important to try to close the file,
                //  otherwise we may run out of file descriptors!
                if (ois != null) {
                    ois.close();
                }
            } catch (IOException e) {
                throw new ObjectIOException(e);
            }
        }
    }

    /**
     * Writes an array of objects with the given id and type to
     * this storage medium, in the chosen directory.
     * 
     * @param node	the directory in which the file should be saved.
     * @param id	the id under which to store the StorableObject.
     * @param objType	the type of StorableObject.
     * @param obj	the array of objects and object references to store,
     *			as described in the section for the read method.
     * 			if an I/O error occurred while reading the file
     * 
     * @exception ObjectIOException
     * 			if an I/O error occurred while writing the file
     * 
     * @see #read
     * @see #update
     */
    public void write(
        DirectoryNode node,
        long id,
        Class objType,
        UnpackedStorableObject obj)
        throws ObjectIOException {

        File file = new Filename(node, id, objType).getFile();
        Filename oldfname = new Filename(node, id, objType);
        Filename newfname = new Filename(node, id, objType);
        ObjectOutputStream oos = null;
        OutputStream os;
        File oldf;
        File newf;

        oldfname.setBackupVersion(true);
        oldf = oldfname.getFile();

        newfname.setNewVersion(true);
        newf = newfname.getFile();

        try {
            os = new FileOutputStream(newf.getAbsolutePath());
            oos = new ObjectOutputStream(new BufferedOutputStream(os));

            oos.writeObject(obj);
            oos.close(); // New file has now been written to newf.

            oos = null; // Set to null, so we don't close it twice.

            oldf.delete(); // Safety: oldf shouldn't exist but...
            file.renameTo(oldf);
            // <-- If we crash here,
            //     recoverLocalStorage() will fix.
            newf.renameTo(file);
            oldf.delete();
        } catch (IOException e) {
            e.printStackTrace();
            throw new ObjectIOException(e);
        } finally {
            try {
                // It's important to try to close the file,
                //  otherwise we may run out of file descriptors!
                if (oos != null) {
                    oos.close();
                }
            } catch (IOException e) {
                throw new ObjectIOException(e);
            }
        }

    }

    /**
     * Updates an array of objects with the given id and type on
     * this storage medium, in the chosen directory.
     * 
     * @param node	the directory in which the file should be saved.
     * @param id	the id under which to store the StorableObject.
     * @param objType	the type of StorableObject.
     * @param obj	the array of objects and object references to store,
     *			as described in the section for the read method.
     * 			if an I/O error occurred while reading the file
     * 
     * @exception ObjectIOException
     * 			if an I/O error occurred while writing the file
     * 
     * @see #read
     * @see #write
     */
    public void update(
        DirectoryNode node,
        long id,
        Class objType,
        UnpackedStorableObject obj)
        throws ObjectIOException {
        write(node, id, objType, obj);
    }

    /**
     * Updates an array of objects with the given id and type on
     * this storage medium, in the chosen directory.
     * 
     * @param node	the directory in which the file should be saved.
     * @param id	the id under which to store the StorableObject.
     * @param objType	the type of StorableObject.
     * 
     * exception ObjectIOException
     * 			if an I/O error occurred while writing the file
     * 
     * @see #read
     * @see #write
     */
    public void delete(DirectoryNode node, long id, Class objType) {
        File file = new Filename(node, id, objType).getFile();

        file.delete();
    }

    /**
     * Moves an object between directories in local storage.
     * 
     * @param id	id of the StorableObject.
     * @param cls	type of StorableObject.
     * @param fromNode	directory from which the object should be moved.
     * @param toNode	directory to which the object should be moved.
     * 
     * @exception ObjectIOException
     * 			if an I/O error occurred while during the operation
     */
    public void move(
        long id,
        Class objType,
        DirectoryNode fromNode,
        DirectoryNode toNode)
        throws ObjectIOException {
        File fromFile = new Filename(fromNode, id, objType).getFile();
        File toFile = new Filename(toNode, id, objType).getFile();

        if (!fromFile.renameTo(toFile)) {
            // We don't know why, but it failed.
            throw new ObjectIOException("Cannot move object");
        }
    }

    /**
     * Builds the DirectoryNode hierarchy based on the directory
     * structure on disk.
     * 
     * @exception ObjectIOException
     *			if there was an I/O error during the operation
     */
    public void initialise() throws ObjectIOException {
        buildDirectoryTree();
        recoverLocalStorage();
    }

    /**
     * Builds the DirectoryNode tree representing the structure of
     * the local storage directories.
     */
    private void buildDirectoryTree() {
        FilenameFilter dirFilter = new FilenameFilter() {
            public boolean accept(File dir, String name) {
                return name.endsWith(DIRNAME_SUFFIX);
            }
        };

        Stack toDo = new Stack();
        toDo.push(subtreeDirFile_);
        toDo.push(getRoot());

        String[] subDirs;
        String dirName;
        DirectoryNode node, child, top;
        File nodeFile;

        while (!toDo.isEmpty()) {
            node = (DirectoryNode) toDo.pop();
            nodeFile = (File) toDo.pop();

            subDirs = nodeFile.list(dirFilter);

            if (subDirs == null) {
                continue;
            }

            OUTER : for (int j = 0; j < subDirs.length; j++) {
                long id = new Filename(node, subDirs[j]).getId();

                // Not found, add node to hierarchy and queue subdirs
                if (!node.hasSubdir(id)) {
                    node.mkdir(id);
                }

                top = node.getDirectory(id);

                toDo.push(new Filename(top).getFile());
                toDo.push(top);
            }
        }
    }

    /**
     * Ensure that the local storage area is in a consistent state.
     * In particular, if the system crashed while writing an object
     * we either reinstate the old version of the file or move
     * the new version into place.
     * 
     * @exception ObjectIOException
     *			if there was an I/O error during the operation
     */
    private void recoverLocalStorage() throws ObjectIOException {
        Enumeration enum = getRoot().nodes();
        DirectoryNode node;
        String[] dirList;
        String fname, suffix;
        File dir;

        // For each directory, check for files to recover.
        while (enum.hasMoreElements()) {
            node = (DirectoryNode) enum.nextElement();
            dir = new Filename(node).getFile();
            dirList = dir.list();

            debug("Recovering:" + dir);

            if (dirList == null)
                continue;

            for (int j = 0; j < dirList.length; j++) {
                fname = dirList[j];
                Filename oldfilename;

                try {
                    oldfilename = new Filename(node, dirList[j]);
                } catch (IllegalArgumentException e) {
                    // Invalid filename, eg random file copied to local storage
                    continue;
                }

                debug("Checking:" + fname);

                suffix = oldfilename.getSuffix();

                // The existence of a .old file indicates a crash.
                if (suffix.equals(OLDFILE_SUFFIX)) {
                    Filename baseFn = new Filename(oldfilename);
                    Filename newFn = new Filename(oldfilename);

                    // Toggle property to lose additional ".old" suffix
                    baseFn.setBackupVersion(false);

                    // Lose additional ".old" suffix, add ".tmp" suffix
                    newFn.setBackupVersion(false);
                    newFn.setNewVersion(true);

                    File oldf = oldfilename.getFile();
                    File baseFile = baseFn.getFile();
                    ;

                    debug("node:" + node);
                    debug("fname:" + fname);
                    debug("base:" + baseFn);
                    debug("base file:" + baseFn.getFile());
                    debug("old:" + oldfilename);
                    debug("old file:" + oldfilename.getFile());
                    debug("new:" + newFn);
                    debug("new file:" + newFn.getFile());

                    if (baseFile.exists()) {
                        // New version was moved into place: delete backup.
                        debug("base file exists:" + baseFile);
                        checkOk(oldf.delete(), "deleting");
                    } else {
                        // There is no base file, such as "123.tp"
                        // Probably "123.tp.tmp" exists and, if so,
                        //   should become the live version.
                        File newf = newFn.getFile();

                        if (newf.exists()) {
                            debug("installing new file:" + newf);
                            // New version exists: rename and remove backup
                            checkOk(newf.renameTo(baseFile), "renaming");
                            checkOk(oldf.delete(), "deleting");
                        } else {
                            debug("reverting to oldf:" + oldf);
                            // Should never happen but...
                            // No new version exists: revert to backup copy
                            checkOk(oldf.renameTo(baseFile), "renaming");
                        }
                    }
                }
            }
        }
    }

    /**
     * Utility method to throw an ObjectIOException if <code>succeeded</code>
     * is false.  Used to check that the results of I/O operations.
     * 
     * @param succeeded		true if everything is okay,
     * 				false if an error occurred.
     * @param description	Description of the error.
     * 
     * @exception ObjectIOException
     *			thrown if <code>succeeded</code> is false,
     * 			usually because some I/O operation failed.
     */
    private void checkOk(boolean succeeded, String description)
        throws ObjectIOException {
        if (!succeeded) {
            throw new ObjectIOException(
                "recovering: error while " + description);
        }
    }

    /**
     * Creates the directory on the file system.
     * @param node The directory where the directory should be placed
     * @param id The id of the directory to be created
     *     
     * @return True if directory is created successfully else false
     */
    public boolean createDirectory(DirectoryNode node, long id) {

        if (node.mkdir(id)) {
            DirectoryNode newNode = node.getDirectory(id);
            File dirFile = new Filename(newNode).getFile();

            if (dirFile.mkdir()) {
                return true;
            } else {
                node.rmdir(id);
            }
        }

        return false;
    }

    /**
     * Moves a directory within the file system.
     *
     * @param node	directory to be moved
     * @param toNode	the new parent directory
     * 
     * @return True on success false on error
     * 
     * @exception ObjectNotFoundException
     *			if the object was not found in fromNode.
     * 
     * @exception ObjectIOException
     *			thrown if some I/O operation fails.
     */
    public boolean moveDirectory(DirectoryNode fromNode, DirectoryNode toNode)
        throws ObjectIOException, ObjectNotFoundException {
        File fromDir = new Filename(fromNode).getFile();
        File toDir =
            new Filename(toNode, new Filename(fromNode).getFilename())
                .getFile();

        if (!fromDir.exists()) {
            throw new ObjectNotFoundException(
                "Not found: " + fromDir.getAbsolutePath());
        }

        if (fromDir.renameTo(toDir)) {
            fromNode.moveDirectory(toNode);
            return true;
        } else {
            return false;
        }
    }

    /**
     * Deletes the directory from the file system.
     *     
     * @param node The directory to be deleted.
     *     
     * @return True if the directory is deleted successfully otherwise false   
     */
    public boolean deleteDirectory(DirectoryNode node) {
        File dirFile = new Filename(node).getFile();

        return dirFile.delete() && node.getParent().rmdir(node.getId());
    }

    /**
     * Builds a list of Summary objects for a given directory.
     * These Summary objects contain the unique id of the object
     * and the Class which contains details about the type of object
     * that is being referenced, be it an ObservationDescription,
     * ObservationBlock etc.. 
     *
     * @param node Directory from which you can create a list of objects
     *
     * @return array of Summary objects
     */
    public Summary[] listObjects(DirectoryNode node) {
        File dirFile = new Filename(node).getFile();
        String filenames[] = dirFile.list();

        Config cfg = Config.getCfg();
        int numObjects = 0;
        Class cls;
        String fname;

        if (filenames == null) {
            return new Summary[] {
            };
        }

        Summary sum[] = new Summary[filenames.length];

        for (int i = 0; i < filenames.length; i++) {
            try {
                Filename fn = new Filename(node, filenames[i]);

                cls = cfg.getClassForFileSuffix(fn.getSuffix());

                if (cls != null && cfg.isUserLevel(cls)) {
                    Summary info = new Summary();

                    sum[numObjects] = info;
                    info.setId(fn.getId());
                    info.setObjectClass(cls);
                    numObjects++;
                }
            } catch (IllegalArgumentException e) {
                // Bad filename, eg random file copied into local storage
                continue;
            }
        }

        Summary[] result = new Summary[numObjects];
        System.arraycopy(sum, 0, result, 0, numObjects);

        return result;
    }

    /**
     * Do a EVM verification test for the specified objects in a single atomic operation.
     *
     * @param   usos    Descriptions of each StorableObject to do EVM verification
     *
     * @return boolean if OB passed EVM test
     */
    public boolean verifyObject(MultiObjSpec[] moi) throws ObjectIOException {
        throw new ObjectIOException("verifyObject: Not implemented for local storage");
    }

    /**
     * Download an IP - this is not applicable for Local Storage, and exception is thrown.
     */
    public byte[] downloadInstrumentPackage(String instrument, float version)
        throws ObjectIOException {
        throw new ObjectIOException("downloadInstrumentPackage: not implemented for local storage.");
    }


/**
     * Download an IP - this is not applicable for Local Storage, and exception is thrown.
     */
    public FindingChart [] populateFindingCharts(long boId)
        throws ObjectIOException {
        throw new ObjectIOException("populateFindingCharts: not implemented for local storage.");
    }

    /**
     * Checkin the readme file for a given observing run
     * @param Readme file business object
     */
    
    public void checkInReadme(ObservingRun obsRun)
    throws ObjectIOException {
        throw new ObjectIOException("checkInReadme: not implemented for local storage.");
    }
    
    /**
     * Checkout the readme file for a given observing run.
     * @param boId Observing run definition. 
     */
    
    public String checkOutReadme(ObservingRun obsRun)
    throws ObjectIOException {
        throw new ObjectIOException("checkOutReadme: not implemented for local storage.");
    }
    
    
    

    /**
     * Check in the specified objects in a single atomic operation.
     * 
     * @param	usos	Descriptions of each StorableObject to be checked in.
     * 
     * @return		array of error strings.
     *			If length = 0, no errors.
     */
    public void checkIn(MultiObjSpec[] moi) throws ObjectIOException {
        throw new ObjectIOException("CheckIn: Not implemented for local storage");
    }

    /**
     * Check out the ObsBlk or CalBlk with the specified id plus
     * any attached objects, ie Target, ObsDescr, etc.
     * 
     * @param	id	id of the ObsBlk or CalBlk to be checked out
     * 
     * @return		CheckOutStatus object that contains a return
     *			status and (if successful) references to the
     *			checked-out objects.
     */
    public void checkOut(IdType[] idTypes)
        throws ObjectNotFoundException, ObjectIOException {
        throw new ObjectIOException("CheckOut: Not implemented for local storage");
    }

    /**
     * Write several object to the external media in an atomic operation.
     *      
     * @param moi	Array of object descriptions.
     * 
     * @exception ObjectIOException
     * 			if an I/O error occurred during the operation
     * 
     * @see #read
     * @see #write
     * @see #update
     */
    public void multiStore(MultiObjSpec[] mos) throws ObjectIOException {

        for (int j = 0; j < mos.length; j++) {
            MultiObjSpec m = mos[j];
            if (m.isNew()) {
                write(m.dir, m.obj.getObject().getId(), m.getCls(), m.obj);
            } else {
                update(m.dir, m.obj.getObject().getId(), m.getCls(), m.obj);
            }
        }
    }

    private void debug(String msg) {
        // stdlog_.debug(msg);
    }

    /**
     * Returns a string representation of this object and its fields.
     * 
     * @return a string representation.
     */
    public String toString() {
        return "LocalDirMgr[subtreeDir="
            + subtreeDir_
            + ",rootNode="
            + getRoot()
            + "]";
    }

    public class Filename {

        /**
         * Unique id for a given filename
         */

        private long id_;

        /**
             * Suffix for a given filename
             */

        private String suffix_;

        /**
         * Parent directory for file
         */

        private DirectoryNode parent_;

        private boolean isNewVersion_;

        private boolean isBackupVersion_;

        /**
         * Constructor for normal files.
         */
        public Filename(Filename fn) {
            id_ = fn.id_;
            suffix_ = fn.suffix_;
            parent_ = fn.parent_;
            isNewVersion_ = fn.isNewVersion_;
            isBackupVersion_ = fn.isBackupVersion_;
        }

        /**
         * Constructor for normal files.
             * @param parent DirectoryNode for file
             * @param id Unique id for filename
             * @param class Class that the file is serializing
         */
        public Filename(DirectoryNode parent, long id, Class cls) {
            parent_ = parent;
            id_ = id;
            suffix_ = BusinessObject.getSuffix(cls);
        }

        /**
         * Constructor for directories.
             * @param dir Directory Node for class file
         */
        public Filename(DirectoryNode dir) {
            parent_ = dir.getParent();
            id_ = dir.getId();
            suffix_ = DIRNAME_SUFFIX;
        }

        /**
         * Constructor for files within a given directory.
         *
             * @param parent parent directory
             * @param filename string
             */
        public Filename(DirectoryNode parent, String filename) {
            parent_ = parent;

            int sign = (filename.charAt(0) == 'L' ? -1 : +1);

            debug("orig filename = " + filename);
            suffix_ = getSuffixOf(filename);
            if (suffix_.equals(OLDFILE_SUFFIX)) {
                isBackupVersion_ = true;
                debug("in OLD case, filename becomes: " + filename);
                filename = withoutSuffix(filename);
                suffix_ = getSuffixOf(filename);
            } else if (suffix_.equals(NEWFILE_SUFFIX)) {
                isNewVersion_ = true;
                filename = withoutSuffix(filename);
                debug("in NEW case, filename becomes: " + filename);
                suffix_ = getSuffixOf(filename);
            }

            String prefix = withoutSuffix(filename);
            debug("prefix = " + prefix);
            debug("suffix = " + suffix_);
            debug("filename = " + filename);

            id_ = sign * new Long(prefix.substring(1)).longValue();
        }

        /**
         * Parant of Directory 
         * @return  Parents DirectoryNode 
         */

        public DirectoryNode getParent() {
            return parent_;
        }

        /**
         * Unique id for the file
        * @return id_ unique id
         */

        public long getId() {
            return id_;
        }

        /**
         * Returns the suffix of the file
         * @param suffix
         */

        public String getSuffix() {
            if (isNewVersion_) {
                return NEWFILE_SUFFIX;
            } else if (isBackupVersion_) {
                return OLDFILE_SUFFIX;
            }
            return suffix_;
        }

        /**
         * Returns the file hadle for the file
         * @return File File handle
         */

        public File getFile() {
            if (parent_ == null) {
                return new File(subtreeDir_);
            }

            Stack stk = new Stack();
            StringBuffer sb = new StringBuffer();
            DirectoryNode dir = parent_;

            while (dir.getParent() != null) {
                stk.push(makeFilename(dir.getId(), DIRNAME_SUFFIX));
                dir = dir.getParent();
            }

            sb.append(subtreeDir_);
            while (!stk.isEmpty()) {
                sb.append(File.separator);
                sb.append(stk.pop());
            }

            return new File(sb.toString(), getFilename());
        }

        /**
         * Get filename
         * @return filename
         */

        public String getFilename() {
            if (parent_ == null) {
                throw new IllegalArgumentException("Cannot get name of root");
            }
            String filename = makeFilename(id_, suffix_);

            if (isNewVersion_) {
                filename += "." + NEWFILE_SUFFIX;
            } else if (isBackupVersion_) {
                filename += "." + OLDFILE_SUFFIX;
            }

            return filename;
        }

        /**
         * Set Backup Version to True
         * @return isBackup Backup Version to true
         */

        public void setBackupVersion(boolean isBackup) {
            isBackupVersion_ = isBackup;
        }

        /**
         * Backup version
         * @return boolen if isBackUpVersion_ = true else false
         */

        public boolean getBackupVersion() {
            return isBackupVersion_;
        }

        /**
         * Set the new version of the file
         * @param isNew New version of the file to true
         */

        public void setNewVersion(boolean isNew) {
            isNewVersion_ = isNew;
        }

        /**
         * Get New Version
         * @return  isNewVersion
         */

        public boolean getNewVersion() {
            return isNewVersion_;
        }

        public String getBaseName() {
            if (parent_ == null) {
                throw new IllegalArgumentException("Cannot get name of root");
            }
            return makeFilename(id_, suffix_);
        }

        /**
         * Get the suffix for the filename
         * @param filename Suffix for the filename
         * @return Filename suffix
         */

        private String getSuffixOf(String filename) {
            int lastDotIndex = filename.lastIndexOf('.');
            return lastDotIndex >= 0
                ? filename.substring(lastDotIndex + 1)
                : "";
        }

        /**
         * Filename without suffix
         * @param filename 
         * @return Filename without suffix
         */

        private String withoutSuffix(String filename) {
            int lastDotIndex = filename.lastIndexOf('.');
            return lastDotIndex >= 0
                ? filename.substring(0, lastDotIndex)
                : filename;
        }

        /**
         * Make a filename with the suffix and the id
         * @param id Id of filename
         * @param suff Suffix for file
         */

        private String makeFilename(long id, String suff) {
            return (id < 0 ? "L" : "G") + Math.abs(id) + "." + suff;
        }

        /**
        * toString file
         * @param return string value
         */

        public String toString() {
            String params =
                "id=" + id_ + ", suffix=" + suffix_ + ", parent=" + parent_;

            if (isNewVersion_) {
                params += ",isNew";
            }

            if (isBackupVersion_) {
                params += ",isBackup";
            }

            return "Filename[" + params + "]";
        }

    }

    /* (non-Javadoc)
     * @see org.eso.ohs.persistence.StorageManager#usdSubmit(org.eso.ohs.dfs.ObservingRun)
     */
    public void usdSubmit(ObservingRun obsRun) throws ObjectIOException {
        // TODO Auto-generated method stub
        throw new ObjectIOException("usdSubmit: not implemented for local storage.");
        
    }

    /* (non-Javadoc)
     * @see org.eso.ohs.persistence.StorageManager#populateReadmeFile(org.eso.ohs.dfs.ObservingRun)
     */
    public Readme populateReadmeFile(ObservingRun obsRun) throws ObjectIOException {
        throw new ObjectIOException("populateReadmeFile is not implemented for local storage");
    }

    /* (non-Javadoc)
     * @see org.eso.ohs.persistence.StorageManager#setReadmeFileStatus(java.lang.String, long)
     */
    public void setReadmeFileStatus(String status, long runId) throws ObjectIOException {
        throw new ObjectIOException("Not implemented for local storage");
    }

    /* (non-Javadoc)
     * @see org.eso.ohs.persistence.StorageManager#populateReadmeFileReplicate(org.eso.ohs.dfs.ObservingRun)
     */
    public Readme populateReadmeFileReplicate(ObservingRun obsRun) throws ObjectIOException {
        throw new ObjectIOException("Not implemented for local storage");

    }

}
