package org.eso.ohs.phase2.orang;

import java.util.Date;
import java.util.Iterator;
import java.util.LinkedList;

import org.eso.ohs.dfs.ObservationBlock;
import org.eso.ohs.dfs.TimeInterval;
import org.eso.ohs.phase2.timeline.TimelineTask;
import org.eso.ohs.phase2.visibility.PCF;
import org.eso.ohs.phase2.visibility.PCFInterval;
import org.eso.ohs.phase2.visibility.VisibilityCalcUtils;

/**
 * This class maintains information regarding an ObservationBlock's visibility 
 * and ranking/schedule information.
 * 
 * Created on Jul 12, 2005
 *  * @author pnunes
 *
 */
public class OBScheduleInfo implements TimelineTask {

	private static final double TIME_INTERVAL_THRESOLD = 0.5;

	/*
	 * ObservationBlock
	 */
	private ObDataContainer ob_;
	
	/*
	 * PCFs
	 */
	private PCF resultPCF_;		
	private PCF airmassPCF_;
	
	private PCF sunDownPCF_;
	private PCF moonDownPCF_;	
	private PCF moonAnglePCF_;
	private PCF fliPCF_;
	
	private PCF execTimePCF_;
	
	private PCF timeIntervalPCF_;
	private PCF siderealTimeIntervalPCF_;	
	private PCF meetsZenithConstraintPCF_;
	
	/** Start date of the period over which computation is performed */
	private Date periodStart_;
    
    /** End date of the period over which computation is performed */
	private Date periodEnd_;
    
    /** Length of the interval for which computation is performed, in minutes */
	private int computeInterval_;
	
    /** Scheduled start time for this observation */
	protected Date schedStart;
    
    /** Scheduled end time for this observation */
    protected Date schedEnd;
    
	/*
	 * Intervals in which is observable
	 */
	private LinkedList observIntervals_;
	
	
	
	
	/*
	 * Ranking score
	 */
	private String rankingScore_;

	
	
	/**
	 * Constructs de object
	 */
	public OBScheduleInfo() {
		observIntervals_ = new LinkedList();
	}	
	
	public OBScheduleInfo( ObservationBlock ob ) {
		this();
		ob_ = new ObDataContainer(ob);
	}
	
	/*
	 * Add an interval in which this OB is observable
	 * @param start
	 * @param end
	 */
	public void addObservableInterval(Date start, Date end) {
		PCFInterval interval = new PCFInterval(start,end,0.0);
		observIntervals_.add(interval);		
	}
	
	/*
	 * Returns an array of Intervals within which the OB is observable
	 * @return PCFInterval [] 
	 */
	public PCFInterval [] getObservableIntervals() {
		PCFInterval [] intervals = new PCFInterval[observIntervals_.size()];
		
		int i = 0;
		for (Iterator iter = observIntervals_.iterator(); iter.hasNext();) {
			PCFInterval interval = (PCFInterval) iter.next();
			intervals[i] = interval;
			i++;
		}		
		return intervals;
	}
	
	
	/*
	 * Setter and getter methods
	 */
	
	public PCF getAirmassPCF() {
		return airmassPCF_;
	}
	public void setAirmassPCF(PCF airmassPCF) {
		this.airmassPCF_ = airmassPCF;
	}
    
    /** @see #periodEnd_ */
	public Date getEnd() {
		return periodEnd_;
	}
    
    /** @see #periodEnd_ */
	public void setEnd( Date end ) {
        this.periodEnd_ = end;
    }
    
    /** @see #computeInterval_ */
	public int getInterval() {
		return computeInterval_;
	}
    
    /** @see #computeInterval_ */
	public void setInterval( int interval ) {
        this.computeInterval_ = interval;
    }
    
	public PCF getMoonAnglePCF() {
		return moonAnglePCF_;
	}
	public void setMoonAnglePCF(PCF moonAnglePCF) {
		this.moonAnglePCF_ = moonAnglePCF;
	}
	public ObDataContainer getOb() {
		return ob_;
	}
	public void setOb(ObDataContainer ob) {
		this.ob_ = ob;
	}
	public boolean isObservable() {
		return ( observIntervals_.size() > 0 );
	}
	
	public PCF getResultPCF() {
		return resultPCF_;
	}
	public void setResultPCF(PCF resultPCF) {
		this.resultPCF_ = resultPCF;
	}
    
    /** @see #periodStart_ */
	public Date getStart() {
		return periodStart_;
	}
    
    /** @see #periodStart_ */
	public void setStart( Date start ) {
        this.periodStart_ = start;
    }

	public PCF getFliPCF() {
		return fliPCF_;
	}
	public void setFliPCF(PCF fliPCF) {
		this.fliPCF_ = fliPCF;
	}
	
	public PCF getExecTimePCF() {
		return execTimePCF_;
	}
	public void setExecTimePCF(PCF execTimePCF_) {
		this.execTimePCF_ = execTimePCF_;
	}
		
	
	public PCF getSiderealTimeIntervalPCF() {
		return siderealTimeIntervalPCF_;
	}
	public void setSiderealTimeIntervalPCF(PCF siderealTimeIntervalPCF) {
		this.siderealTimeIntervalPCF_ = siderealTimeIntervalPCF;
	}
	public PCF getTimeIntervalPCF() {
		return timeIntervalPCF_;
	}
	public void setTimeIntervalPCF(PCF timeIntervalPCF) {
		this.timeIntervalPCF_ = timeIntervalPCF;
	}
	public PCF getMoonDownPCF() {
		return moonDownPCF_;
	}
	public void setMoonDownPCF(PCF moonUpPCF) {
		this.moonDownPCF_ = moonUpPCF;
	}
	public PCF getSunDownPCF() {
		return sunDownPCF_;
	}
	public void setSunDownPCF(PCF sunUpPCF) {
		this.sunDownPCF_ = sunUpPCF;
	}

	public String getRankingScore() {
		return rankingScore_;
	}

	public void setRankingScore(String rankingScore) {
		this.rankingScore_ = rankingScore;
	}	
	
	/**
	 * Determine the part of the interval in which the date falls. 
	 * Returns 0 if it is the beginning and 1 if it is in the end of the interval 
	 * @return a double between 0 and 1 if the date falls within a time interval, otherwise returns -1
	 */
	public double whereInInterval(Date now) {
		
		double res = -1;
		
		TimeInterval [] intervals = ob_.getTimeIntervals();						
		for (int i = 0; i < intervals.length; i++) {
			TimeInterval ti = intervals[i];
			
			Date startTime =
                new Date(ti.getStartTime() * VisibilityCalcUtils.MILLISECS_PER_SEC);
            Date endTime =
                new Date(ti.getEndTime() * VisibilityCalcUtils.MILLISECS_PER_SEC);
            
            if ( startTime.getTime() <= now.getTime() )
            	if (now.getTime() <= endTime.getTime()) {
            		
            		double a =  now.getTime()-startTime.getTime();
            		double b = endTime.getTime()-startTime.getTime();
            		
            		res = a/b;            		
            	}
		}
		
		return(res);
		
	}
	
	public boolean isSecondHalfTimeInterval() {
		
		double critical = whereInInterval(periodStart_);
		
		if ( critical > TIME_INTERVAL_THRESOLD )
			return(true);
		else
			return(false);
	}
	
	
	public String toString() {
		StringBuffer buf = new StringBuffer();
		buf.append("OB #");
		buf.append(getOb().getObInfo().getId());
		buf.append("\n");
		
		
		buf.append("-- FLI: \n");
		buf.append(getFliPCF().toStringCommas());
		buf.append("-- FLI: \n");
		
		buf.append("-- Moon Angle: \n");
		buf.append(getMoonAnglePCF().toStringCommas());
		buf.append("-- Moon Angle: \n");
		
		buf.append("-- Moon Down: \n");
		buf.append(getMoonDownPCF().toStringCommas());
		buf.append("-- Moon Down: \n");
		
		buf.append("-- Sun down: \n");
		buf.append(getSunDownPCF().toStringCommas());
		buf.append("-- Sun down: \n");
		
		buf.append("-- TI: \n");
		buf.append(getTimeIntervalPCF().toStringCommas());
		buf.append("-- TI: \n");
		
		buf.append("-- STI: \n");
		buf.append(getSiderealTimeIntervalPCF().toStringCommas());
		buf.append("-- STI: \n");
		
		buf.append("-- Airmass: \n");
		buf.append(getSunDownPCF().toStringCommas());
		buf.append("-- Airmass: \n");
		
		buf.append("-- Result Observable intervals: " );
		buf.append(observIntervals_.size());
		buf.append("\n");
		buf.append(getResultPCF().toStringCommas());
		buf.append("-- Result: \n");
		
		return(buf.toString());
	}
	
	

	/**
	 * @param meetsZenithConstraintPCF
	 */
	public void setMeetsZenithConstraintPCF(PCF meetsZenithConstraintPCF) {
		this.meetsZenithConstraintPCF_ = meetsZenithConstraintPCF;
		
	}
	

    /**
     * @return zenith constraint PCF
     */
    public PCF getMeetsZenithConstraintPCF() {
		return meetsZenithConstraintPCF_;
	}

	/**
     * @see TimelineTask#getDuration()
     */
    // amchavan, 20-Sep-2005: added
    public long getDuration() {
        long t = getOb().getObInfo().getExecutionTime();
        return t;
    }

    /**
     * @see TimelineTask#getID()
     */
    // amchavan, 20-Sep-2005: added
    public long getID() {
        return getOb().getObInfo().getDbaseId();
    }

    /**
     * @see TimelineTask#getPreference()
     */
    // amchavan, 20-Sep-2005: added
    public PCF getPreference() {
        return getResultPCF();
    }

    /**
     * @see TimelineTask#getStartTime()
     */
    // amchavan, 20-Sep-2005: added
    public Date getStartTime() {
        return this.schedStart;
    }

    /**
     * @see TimelineTask#setStartTime(Date)
     */
    // amchavan, 20-Sep-2005: added
    public void setStartTime( Date start ) {
        this.schedStart = start;
        if( start != null )
            this.schedEnd = new Date( start.getTime() + getDuration() * 1000 );
    }

    /**
     * @see TimelineTask#getName()
     */
    // amchavan, 20-Sep-2005: added
    public String getName() {
        return this.ob_.getObInfo().getName();
    }

    /**
     * @see TimelineTask#compareTo(TimelineTask)
     */
    // amchavan, 20-Sep-2005: added
    public int compareTo( TimelineTask anotherTask ) {
        OBScheduleInfo other = (OBScheduleInfo) anotherTask;
        return getRankingScore().compareTo( other.getRankingScore() );
    }

    /**
     * @see TimelineTask#isDoable()
     */
    // amchavan, 20-Sep-2005: added
    public boolean isDoable() {
        return isObservable();
    }

    /**
     * @see TimelineTask#getEndTime()()
     */
    // amchavan, 12-Oct-2005: added
    public Date getEndTime() {
        return this.schedEnd;
    }
    
    /**
     * The equals checks if an object has the same obId properties
     * 
     * 
     * @see java.lang.Object#equals(java.lang.Object)
     */
    public boolean equals(Object obj) {
    	boolean result = false;
    	if (obj!= null&& obj instanceof OBScheduleInfo) {
    		OBScheduleInfo currentOb = (OBScheduleInfo) obj;
    		result = (this.getID() == currentOb.getID());
    	}
    	return result;
    } 
   
}
