package org.eso.ohs.phase2.visiplot;

/**
 * VisiPlot.java
 *
 * Created on 1 August 2001 13:00 PM
 *
 * @version $Id: OBVisiplotInfo.java,v 1.6 2005/11/18 17:56:54 pnunes Exp $
 * @author cmavros  for ESO
 * <br>ChangeLog: cmavros 24 Apr 02 - 1722/OT 
 */

import java.util.Date;

import org.apache.log4j.Logger;
import org.eso.ohs.dfs.ObservationBlock;
import org.eso.ohs.phase2.orang.OBScheduleInfo;
import org.eso.ohs.phase2.visibility.PCF;
import org.eso.ohs.phase2.visibility.PCFInterval;
import org.eso.ohs.phase2.visibility.VisibilityCalcUtils;

public class OBVisiplotInfo {

    /** Standard log channel 
     * <br>ChangeLog: cmavros 21-12-01 added
     */
    private static Logger stdlog_ = Logger.getLogger(OBVisiplotInfo.class);

    /**
     * Constraint values
     */
    private double airConstraint_;
    private double fliConstraint_;
    private double moonConstraint_;
    private double sunConstraint_;

    private Date startTime_;
    private Date endTime_;
    private double execTime_;
    private int interval_;

    /**
     * VisiPlot Options 
     */
    private boolean airEnabled_;
    private boolean fliEnabled_;
    private boolean moonEnabled_;
    private boolean sunEnabled_;
    private boolean execTimeEnabled_;
    private boolean timeIntervalEnabled_;
    private boolean siderealTimeIntervalEnabled_;

    /**
     * Do not care about moon angle or FLI if the moon is not up
     * this option indicates if the moon up/down status should be 
     * considered or not (currently set to true, but maybe a configuration
     * option later).
     */
    private boolean moonVisibilityEnabled_ = true;

    /**
     * PCF Plotter values
     */
    private PCF fliPCF_;
    private PCF moonDistPCF_;
    private PCF airMassPCF_;
    private PCF sunDownPCF_;
    private PCF moonDownPCF_;
    

    /**
     * PCFs 
     */
    private PCF resultPCF_;
    private PCF timeIntervalPCF_;
    private PCF siderealTimeIntervalPCF_;
    private PCF execTimePCF_;
  
        
   
    /**
     * Create and calculate PCFs.
     *
     * @param current ObservationBlock
     * @param current VisiPlot Options 
     */
    public OBVisiplotInfo(OBScheduleInfo ob, VisiPlotOptions vpo) {    	
        readVisiPlotOptions(ob, vpo); 
        setPCFs(ob);  
    }
    
    
    /*
     * Create a PCF taking an OBScheduleInfo object as a parameter. It doesn't calculate
     * the PCFs.
     */    
    public OBVisiplotInfo(OBScheduleInfo ob) {
    	setPCFs(ob);
    }
    
    
    /**
     * Adds some changes to the PCFs, required by plotting
     * Basically adds a positive infinity point at the end
     * @param ob
     */
    private void setPCFs( OBScheduleInfo ob ) {
    	
    	airMassPCF_ = new PCF();
    	addPCFToPCF(ob.getAirmassPCF(),airMassPCF_);
    	airMassPCF_.addInterval(PCF.PositiveInfinity,0);
    	
    	
    	if( fliEnabled_) {
    		fliPCF_ = new PCF();
    		addPCFToPCF(ob.getFliPCF(),fliPCF_);
    		fliPCF_.addInterval(PCF.PositiveInfinity,0);
    	}
    	
    	if ( moonVisibilityEnabled_) {
    		moonDownPCF_ = new PCF();
    		addPCFToPCF(ob.getMoonDownPCF(),moonDownPCF_);
    		moonDownPCF_.addInterval(PCF.PositiveInfinity,0);
    	}
    	
    	if ( moonEnabled_) {
    		moonDistPCF_ = new PCF();
    		addPCFToPCF(ob.getMoonAnglePCF(),moonDistPCF_);
    		moonDistPCF_.addInterval(PCF.PositiveInfinity,0);
    	}
    	
    	
    	if ( sunEnabled_) {
    		sunDownPCF_ = new PCF();    	
    		addPCFToPCF(ob.getSunDownPCF(),sunDownPCF_);
    		sunDownPCF_	.addInterval(PCF.PositiveInfinity,0);
    	}
         
    	if (siderealTimeIntervalEnabled_) {
    		siderealTimeIntervalPCF_ = new PCF();    	
    		addPCFToPCF(ob.getSiderealTimeIntervalPCF(),siderealTimeIntervalPCF_ );        
    		siderealTimeIntervalPCF_	.addInterval(PCF.PositiveInfinity,0);
    	}
        
    	if ( timeIntervalEnabled_) {
    		timeIntervalPCF_ = new PCF();
    		addPCFToPCF(ob.getTimeIntervalPCF(),timeIntervalPCF_ );
    		timeIntervalPCF_.addInterval(PCF.PositiveInfinity,0);
    	}
  	
    	resultPCF_ = ob.getResultPCF();
    	execTimePCF_ = new PCF();
    	addPCFToPCF(ob.getExecTimePCF(),execTimePCF_);
    	execTimePCF_.addInterval(PCF.PositiveInfinity,0);


    }
    
    /** 
     * Utility method
     * @param pcf
     * @param second
     */
    private void addPCFToPCF( PCF pcf , PCF second ) {
        if (pcf!= null) {
    	for (int i = 0; i < pcf.getIntervalCount(); i++) {
			PCFInterval interval = pcf.getIntervalAt(i);
    		
			if (i == 0) {
				second.addInterval(
                    interval.getStart(),
                    interval.getEnd(),
                    interval.getValue());
            } else {
            	second.appendInterval(
                		interval.getEnd(),interval.getValue());
            }
		}
        }
    	
    }
    

    /**
     * <br>ChangeLog: cmavros 24 Apr 02  - 1722/OT
     * @return time interval pcf
     */
    public PCF getTimeIntervalPCF() {
        return (timeIntervalPCF_);
    }

    /**
     * <br>ChangeLog: cmavros 24 Apr 02  - 1722/OT
     * @return time interval pcf
     */
    public PCF getSiderealTimeIntervalPCF() {
        return (siderealTimeIntervalPCF_);
    }

    /**
     * @return exec time pcf
     */
    public PCF getExecTimePCF() {
        return (execTimePCF_);
    }

    /**
     * @return fli pcf
     */
    public PCF getFLIPCF() {
        return (fliPCF_);
    }

    /**
     * @return moon dist pcf
     */
    public PCF getMoonDistPCF() {
        return (moonDistPCF_);
    }

    /**
     * @return PCF of moon down status (1.0: its down, 0.0: its up).
     */
    public PCF getMoonDownPCF() {
        return (moonDownPCF_);
    }

    /**
     * @return the start time of the plot
     */
    public Date getStartTime() {
        return startTime_;
    }

    /**
     * @return the end time of the plot 
     */
    public Date getEndTime() {
        return endTime_;
    }

    /**
     * @return the precison of the plot
     */
    public int getPrecision() {
        return interval_;
    }

    /**
     * @return true if AirMass is enabled 
     */
    public boolean isAirMassEnabled() {
        return airEnabled_;
    }

    /**
     * @return air mass pcf
     */
    public PCF getAirMassPCF() {
        return (airMassPCF_);
    }

    /**
     * @return sun pcf
     */
    public PCF getSunPCF() {
        return (sunDownPCF_);
    }
    
    /**
     * @return result PCF_;
     */
    public PCF getResultPCF() {
    	return(resultPCF_);
    }

    /**
     * @return time interval 
     */
    public double getTimeInterval(ObservationBlock ob, Date date) {
        return (VisibilityCalcUtils.getTimeIntervalValue(ob.getTimeIntervals(), date));
    }

   
    /**
     * readVisiPlotOptions gets the user selected options
     * from the VisiPlot selection options.
     * 
     * @param current ObservationBlock
     * @param current VisiPlot Options 
     */
    private void readVisiPlotOptions(
    	OBScheduleInfo ob,
        VisiPlotOptions vpo) {
        String val = null;

        interval_ = vpo.getIntValue(VisiPlotOptions.dataQueryIntKey);
        airEnabled_ = vpo.isEnabled(VisiPlotOptions.airMassConEnabledKey);
        endTime_ = vpo.getISODate(VisiPlotOptions.endTimeKey);
        execTimeEnabled_ = vpo.isEnabled(VisiPlotOptions.obexecConKey);
        fliEnabled_ = vpo.isEnabled(VisiPlotOptions.fliConEnabledKey);
        moonEnabled_ = vpo.isEnabled(VisiPlotOptions.moonDistConEnabledKey);
        startTime_ = vpo.getISODate(VisiPlotOptions.startTimeKey);        
        sunEnabled_ = vpo.isEnabled(VisiPlotOptions.sunConEnabledKey);
        timeIntervalEnabled_ =
            vpo.isEnabled(VisiPlotOptions.obtimeIntervalConKey);

        siderealTimeIntervalEnabled_ =
            vpo.isEnabled(VisiPlotOptions.obSiderealTimeIntervalConKey);

        val = vpo.getValue(VisiPlotOptions.airMassConKey);
        airConstraint_ = getAirMassConstraint(ob, vpo, val);

        val = vpo.getValue(VisiPlotOptions.moonDistConKey);
        moonConstraint_ = getMoonConstraint(ob, vpo, val);

        val = vpo.getValue(VisiPlotOptions.fliConKey);
        fliConstraint_ = getFLIConstraint(ob, vpo, val);

        val = vpo.getValue(VisiPlotOptions.sunConKey);
        sunConstraint_ = getSunConstraint(vpo, val);


        execTime_ = ob.getOb().getObInfo().getExecutionTime();
        
    }

    

    /**
     * getAirMassConstraint calls isReadFromOB()
     * to determines which constraint value to use; 
     * the VisiPlot Options value or the constraint 
     * from the database.
     *
     * @return AirMass Constraint 
     * @see #readVisiPlotOptions(ObservationBlock ob, VisiPlotOptions vpo)
     * @see VisiPlotOptions#isReadFromOB(String value)
     */
    private double getAirMassConstraint(
    	OBScheduleInfo ob,
        VisiPlotOptions vpo,
        String value) {
        airConstraint_ = 0.;
        if (airEnabled_) {
            boolean readFromOB = VisiPlotOptions.isReadFromOB(value);
            if (!readFromOB) {
                try {
                    airConstraint_ = Double.valueOf(value).doubleValue();
                } catch (NumberFormatException e) {
                    airConstraint_ = 0.;
                }
            } else {
                airConstraint_ = ob.getOb().getObInfo().getAirmass();
            }
        }
        //stdlog_.info("visiplot.Visiplot.getAirMassConstraint " + airConstraint_);  
        return (airConstraint_);
    }

    /**
     * getFLIConstraint calls isReadFromOB()
     * to determines which constraint value to use; 
     * the VisiPlot Options value or the constraint 
     * from the database.
     *
     * @return FLI Constraint 
     * @see #readVisiPlotOptions(ObservationBlock ob, VisiPlotOptions vpo)
     * @see VisiPlotOptions#isReadFromOB(String value)
     */
    private double getFLIConstraint(
    		OBScheduleInfo ob,
        VisiPlotOptions vpo,
        String value) {
    	
        boolean readFromOB = VisiPlotOptions.isReadFromOB(value);
        if (!readFromOB) {
            try {
                fliConstraint_ = Double.valueOf(value).doubleValue();
            } catch (NumberFormatException e) {
                fliConstraint_ = 0.;
            }
        } else {
            fliConstraint_ = ob.getOb().getObInfo().getFractionalLunarIllumination();
        }
        //System.out.print(" visiplot.Visiplot.getFLIConstraint " + fliConstraint_);  
        return (fliConstraint_);
    }

    /**
     * getSunConstraint calls isReadFromOB()
     * to determines if the default value of -18 or
     * the VisiPlot Options value is used
     *
     * @return Sun Constraint
     * @see #readVisiPlotOptions(ObservationBlock ob, VisiPlotOptions vpo)
     * @see VisiPlotOptions#isReadFromOB(String value)
     */
    private double getSunConstraint(VisiPlotOptions vpo, String value) {
        boolean readFromOB = VisiPlotOptions.isReadFromOB(value);
        if (!readFromOB) {
            try {
                sunConstraint_ = Double.valueOf(value).doubleValue();
            } catch (NumberFormatException e) {
                sunConstraint_ = -18.;
            }
        } else {
            sunConstraint_ = -18.;
        }
        //System.out.print(" visiplot.Visiplot.getSunConstraint " + sunConstraint_);
        return (sunConstraint_);
    }

    /**
     * getMoonDistance calls isReadFromOB()
     * to determines which constraint value to use; 
     * the VisiPlot Options value or the constraint 
     * from the database.
     *
     * @return Moon Distance Constraint 
     * @see #readVisiPlotOptions(ObservationBlock ob, VisiPlotOptions vpo)
     * @see VisiPlotOptions#isReadFromOB(String value)
     */
    private double getMoonConstraint(
    		OBScheduleInfo ob,
        VisiPlotOptions vpo,
        String value) {
    	
        boolean readFromOB = VisiPlotOptions.isReadFromOB(value);
        if (!readFromOB) {
            try {
                moonConstraint_ = Double.valueOf(value).doubleValue();
            } catch (NumberFormatException e) {
                moonConstraint_ = 0.;
            }
        } else {
            moonConstraint_ = ob.getOb().getObInfo().getMoonAngularDistance();
        }
        //stdlog_.info(" visiplot.Visiplot.getMoonConstraint " + moonConstraint_ + " " + value);
        return (moonConstraint_);
    }

}
