package org.eso.ohs.scripting;

/**
 * Implementation of the ScriptingInterpreter interface for TCL. Using the
 * Jacl script interpreter.
 *
 * The static initializer will register a singleton instance of this class
 * on application startup.
 *
 * An instance to this object should only be obtained through the
 * ScriptInterpreterManager object.
 *
 * @author Nick Kornweibel
 */

import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.LineNumberReader;

import tcl.lang.Interp;
import tcl.lang.TCL;
import tcl.lang.TclException;

public class JaclScriptInterpreter implements ScriptInterpreter {

    /** singleton instance of the JaclScriptInterpreter */
    private static JaclScriptInterpreter instance;

    /** languages supported */
    private static final String[] langs = {"TCL"};

    /** file extensions supported */
    private static final String[] exts = {"tcl"};

    /** single instance of the Jacl interpreter */
    private static Interp interpreter = null;

    /**
     * Static initializer that is executed when class is initialized
     */

    static {
        if (getInterpreter() != null) {
            instance = new JaclScriptInterpreter();
            ScriptInterpreterManager.getInstance()
                     .registerScriptInterpreter(instance);
        }
    }

    /**
     * Private default constructor 
     */

    private JaclScriptInterpreter() {}

    private static Interp getInterpreter() {
        if (interpreter == null) {
            try {
                Class.forName("tcl.lang.Interp");
                interpreter = new Interp();
            } catch (ClassNotFoundException ex) {
                interpreter = null;
            }
        }
        return interpreter;
    }

    /**
     * Execute the script text passed
     * @param script The text to be executed
     */

    public void executeScript(String script)
            throws ScriptInterpreterException {
        try {
            getInterpreter().eval(script);
            
        } catch (TclException ex) {
           
            String exMsg;
            int compCode = ex.getCompletionCode();
            if (compCode == TCL.ERROR) {
                // exMsg = getInterpreter().getResult().toString();
                try {
                    exMsg = getInterpreter().getVar("errorInfo", null, TCL.GLOBAL_ONLY).toString();
                } catch (TclException tclex) {
                    exMsg = "errorInfo variable not found, "+getInterpreter().getResult().toString();
                }
            } else {
                exMsg = ex.toString();
            }
            throw (new ScriptInterpreterException(exMsg));
        }
    }

    /**
     * Execute the text in the file given
     * @param scriptPath The path to the text file
     */

    public void executeFile(File scriptPath)
            throws ScriptInterpreterException {
        /***********************************************************
          The following has been added to work around problems on the Mac
         ************************************************************/                
        StringBuffer scriptBuffer = new StringBuffer();
        try {
            LineNumberReader rdr
                = new LineNumberReader(new FileReader(scriptPath));
            String line = null;
            while ((line = rdr.readLine()) != null) {
                scriptBuffer.append(line+"\n");
            }
        } catch (IOException e) {
            throw (new ScriptInterpreterException(e.toString()));
        }
        executeScript(scriptBuffer.toString());
        
        /***********************************************************
          Original code

        try {
            getInterpreter().evalFile(scriptPath.getPath());
        } catch (TclException ex) {
            String exMsg = null;
            int compCode = ex.getCompletionCode();
            if (compCode == TCL.ERROR) {
                // exMsg = getInterpreter().getResult().toString();
                try {
                    exMsg = getInterpreter().getVar("errorInfo", null, TCL.GLOBAL_ONLY).toString();
                } catch (TclException tclex) {
                    exMsg = "errorInfo variable not found, "+getInterpreter().getResult().toString();
                }
            }
            if (exMsg == null || exMsg.length() == 0) {
                exMsg = ex.toString();
            }
            throw (new ScriptInterpreterException(exMsg));
        }
        *****************************************************************/
    }


    /**
     * Get the names of the languages supported by this interpreter
     * @return Array of String objects of the supported language names
     */

    public String[] getLanguageNames() {
        return langs;
    }

    /**
     * Get the file extensions supported by this interpreter
     * @return Array of String objects of the supported file extensions
     */

    public String[] getFileExtensions() {
        return exts;
    }
}
